#!/usr/bin/env python3
import serial
import sys
import os
import argparse

dev = serial.Serial("/dev/ttyUSB0", 921600)

parser = argparse.ArgumentParser(description='Host-side script for SPHINCS-256 signing')
parser.add_argument('message', metavar='message-file',
                    help='file that contains the message')
parser.add_argument('-u', '--unscramble',
                    help='unscramble the resulting signature',
                    action='store_true', dest='unscramble')
parser.add_argument('-m', '--with-message',
                    help='append the message to the signature',
                    action='store_true', dest='with_message')
args = parser.parse_args()

mlen = os.path.getsize(args.message)

print("Waiting for request..", file=sys.stderr)
dev.read(1)
print("Writing message length {}".format(mlen), file=sys.stderr)
dev.write(int.to_bytes(mlen, length=4, byteorder='little'))
print("Wrote size", file=sys.stderr)

for _ in range(2):
    print("Writing file..", file=sys.stderr)
    with open(args.message, "rb") as m_file:
        for _ in range(64, mlen, 64):
            dev.read(1)
            dev.write(m_file.read(64))
        dev.read(1)
        dev.write(m_file.read())

print("Reading scrambled signature..", file=sys.stderr)
if args.unscramble:
    print("Unscrambling while writing..", file=sys.stderr)

    # originates from api.h for SPHINCS-256
    MESSAGE_HASH_SEED_BYTES = 32
    HASH_BYTES = 32
    TOTALTREE_HEIGHT = 60
    HORST_SIGBYTES = 13312
    HORST_SKBYTES = 32
    WOTS_SIGBYTES = 2144
    N_LEVELS = 12
    SUBTREE_HEIGHT = TOTALTREE_HEIGHT // N_LEVELS

    assert(HASH_BYTES == HORST_SKBYTES)  # this makes unscrambling easier

    sys.stdout.buffer.write(dev.read(MESSAGE_HASH_SEED_BYTES))  # R
    sys.stdout.buffer.write(dev.read((TOTALTREE_HEIGHT+7) // 8))  # leafidx

    # unscrambling
    horstsig = bytearray(HORST_SIGBYTES)
    for _ in range(HORST_SIGBYTES // HASH_BYTES):
        sigpos = int.from_bytes(dev.read(2), byteorder='big')
        horstsig[sigpos:sigpos+HASH_BYTES] = dev.read(HASH_BYTES)
    sys.stdout.buffer.write(horstsig)

    for _ in range(N_LEVELS):
        sys.stdout.buffer.write(dev.read(WOTS_SIGBYTES))
        sys.stdout.buffer.write(dev.read(SUBTREE_HEIGHT*HASH_BYTES))
else:
    sys.stdout.buffer.write(dev.read(41832))

if args.with_message:
    with open(args.message, "rb") as m_file:
        sys.stdout.buffer.write(m_file.read(mlen))

print("Done!", file=sys.stderr)
print("Any further output will now print to stderr:", file=sys.stderr)
while True:
    sys.stderr.buffer.write(dev.read(1))
    sys.stderr.flush()
