/*
hash.c version 20151006
Andreas Hülsing / Joost Rijneveld
Public domain.
*/

#include "prg.h"

#include <stddef.h>

#include "blake256/ref/hash.h"
#include "blake512/ref/hash.h"

extern void chacha_perm_asm(unsigned char* out, const unsigned char* in);

#define SET_KEY_BIT(a,b) (a[15] = (a[15] & 253) | (b << 1))
#define SET_BLOCK_BIT(a,b) (a[15] = (a[15] & 254) | b)

#define WOTS_SELECT_KEY(a) (a[15] = (a[15] & 253) | 1)
#define WOTS_SELECT_BLOCK(a) (a[15] = (a[15] & 254) | 0)

/**
 * Computes an HMAC using BLAKE256 or BLAKE512
 */
void HMAC(unsigned char *out, const unsigned char *in, const size_t inlen, const unsigned char *key, const size_t keylen) {
  // the block size of BLAKE256 / BLAKE512 is 64 and 128, respectively
  unsigned char buf[keylen * 2 + (inlen > keylen ? inlen : keylen)];
  int i = 0;
  memcpy(buf, key, keylen);
  memset(buf+keylen, 0, keylen);
  for (i = 0; i < 2*keylen; i++) {
    buf[i] ^= 0x36;
  }
  memcpy(buf+keylen*2, in, inlen);
  if (keylen == 32) {
    crypto_hash_blake256(buf+2*keylen, buf, keylen * 2 + inlen);
    memcpy(buf, key, keylen);
    memset(buf+keylen, 0, keylen);
    for (i = 0; i < 2*keylen; i++) {
      buf[i] ^= 0x5c;
    }
    crypto_hash_blake256(out, buf, keylen * 3);
  }
  else if (keylen == 64) {
    crypto_hash_blake512(buf+2*keylen, buf, keylen * 2 + inlen);
    memcpy(buf, key, keylen);
    memset(buf+keylen, 0, keylen);
    for (i = 0; i < 2*keylen; i++) {
      buf[i] ^= 0x5c;
    }
    crypto_hash_blake512(out, buf, keylen * 3);
  }
}

/**
 * Implements PRF_m
 */
int prf_m(unsigned char *out, const unsigned char *in, size_t inlen, const unsigned char *key, int keylen)
{
  unsigned int length;
  if (keylen == 32) {
    HMAC(out, in, inlen, key, keylen);
    return 0;
  }
  else if(keylen == 64) {
    HMAC(out, in, inlen, key, keylen);
    return 0;
  }
  return 1;
}

/*
 * Implemts H_m
 */
int hash_m(unsigned char *out,const unsigned char *in,unsigned long long inlen,const unsigned char *key, const int keylen, const int m)
{
  unsigned long long i;
  unsigned char buf[inlen +keylen+m];
  for(i=0;i<m;i++)
  {
    buf[i] = 0x00;
  }
  for(i=0;i <keylen;i++)
  {
    buf[m+i] = key[i];
  }
  for(i=0;i <inlen;i++)
  {
    buf[m+keylen+i] = in[i];
  }
  
  if(m == 32)
  {  
    crypto_hash_blake256(out, buf, inlen +keylen+m);
    return 0;
  } 
  else
  {
    if(m == 64)
    {
      crypto_hash_blake512(out, buf, inlen +keylen+m);
      return 0;
    }
  }
  return 1;
}

static const char *hashc = "expand 32-byte to 64-byte state!";

/**
 * We assume the left half is in in[0]...in[n-1]
 */
int hash_2n_n(unsigned char *out,const unsigned char *in, const unsigned char *pub_seed, unsigned char addr[16], const int n)
{
  
  unsigned char buf[2*n];
  unsigned char key[n];
  unsigned char bitmask[2*n];
  int i;
  
  SET_KEY_BIT(addr,1);
  SET_BLOCK_BIT(addr,0);
  prg_with_counter(key, n, pub_seed, n, addr);
  SET_KEY_BIT(addr,0);
  // Use MSB order
  prg_with_counter(bitmask, n, pub_seed, n, addr);
  SET_BLOCK_BIT(addr,1);
  prg_with_counter(bitmask+n, n, pub_seed, n, addr);
  for(i=0;i<n;i++)
  {
    buf[i] = key[i];
    buf[n+i] = hashc[i];
  }
  chacha_perm_asm(buf, buf);
  for(i=0;i<n;i++)
    buf[i] = buf[i] ^ (in[i] ^ bitmask[i]);
  chacha_perm_asm(buf, buf);
  for(i=0;i<n;i++)
    buf[i] = buf[i] ^ (in[n+i] ^ bitmask[n+i]);
  chacha_perm_asm(buf, buf);
  for(i=0;i<n;i++)
    out[i] = buf[i];
  return 0;
}

int hash_n_n(unsigned char *out,const unsigned char *in, const unsigned char *pub_seed, unsigned char addr[16], const int n)
{ 
  unsigned char buf[2*n];
  unsigned char key[n];
  unsigned char bitmask[n];
  int i;
  
  WOTS_SELECT_KEY(addr);
  prg_with_counter(key, n, pub_seed, n, addr);
  WOTS_SELECT_BLOCK(addr);
  prg_with_counter(bitmask, n, pub_seed, n, addr);
  for(i=0;i<n;i++)
  {
    buf[i] = key[i];
    buf[n+i] = hashc[i];
  }
  chacha_perm_asm(buf, buf);
  for(i=0;i<n;i++)
    buf[i] = buf[i] ^ (in[i] ^ bitmask[i]);
  chacha_perm_asm(buf, buf);
  for(i=0;i<n;i++)
    out[i] = buf[i];
  return 0;
}
